COL_TARGET_AREA = "#FFF0D9"
COL_FIELD_AREA_POSITIONS = None
COL_FIELD_AREA = None
COL_CENTER_OF_FIELD_AREA = None
COL_CENTER_OF_MASS = None
COL_BACKGROUND = None
COL_DEFAULT_OBJECT = "darkgreen"
OPACITY_OBJECT = 1
OPACITY_GUIDES = 0.5